"""Reporter logic to read the scanner state periodically and send it to the backend."""
import psutil

from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis import add_scanner_state
from ostorlab.utils import defintions


class ScannerStateReporter:
    """Reporter collects information about the current scanner."""

    def __init__(
        self,
        scanner_id: str,
        hostname: str,
        ip: str,
    ):
        self._scanner_id = scanner_id
        self.scan_id = None
        self._hostname = hostname
        self._ip = ip
        self.errors = None

    def _capture_state(self) -> defintions.ScannerState:
        """Capture current scanner state."""
        state = defintions.ScannerState(
            scanner_id=self._scanner_id,
            scan_id=self.scan_id,
            cpu_load=psutil.cpu_percent(interval=1, percpu=False),
            total_cpu=psutil.cpu_count(),
            memory_load=psutil.virtual_memory().percent,
            total_memory=psutil.virtual_memory().total >> 30,  # total memory in GB
            hostname=self._hostname,
            ip=self._ip,
            errors=self.errors,
        )
        return state

    def _report_state(self, state: defintions.ScannerState) -> None:
        runner = authenticated_runner.AuthenticatedAPIRunner()
        _ = runner.execute(add_scanner_state.AddScannerStateAPIRequest(state=state))

    async def report(self) -> None:
        """Capture the current state of the scanner and persist it."""
        state = self._capture_state()
        self._report_state(state)



"""Module Responsible for subscribing to nats for different subjects."""
import logging
import asyncio
import functools

from nats.js import errors


from ostorlab.apis import scanner_config
from ostorlab.apis.runners import authenticated_runner
from ostorlab.scanner import handler as scanner_handler
from ostorlab.scanner import nats_conf
from ostorlab.scanner import callbacks
from ostorlab.utils import scanner_state_reporter

WAIT_SCHEDULE_SCAN = 60  # seconds


logger = logging.getLogger(__name__)


def _handle_exception(context):
    """The exception handler for the event loop."""
    logger.error("Caught Loop Exception: %s", context)


class ScanHandler:
    """Class responsible for handling the subscription to bus handler."""

    def __init__(self, state_reporter: scanner_state_reporter.ScannerStateReporter):
        self._bus_handlers = []
        self._state_reporter = state_reporter

    async def close(self) -> None:
        for handler in self._bus_handlers:
            await handler.close()

    async def _subscribe(
        self,
        config: nats_conf.ScannerConfig,
        subject,
        queue,
        cb,
        stream,
        start_at="first",
    ) -> None:
        bus_handler = await self._create_bus_handler(config)
        asyncio.create_task(bus_handler.ensure_running_handler())
        await bus_handler.connect()
        await bus_handler.add_stream(name=stream, subjects=[subject])
        await bus_handler.subscribe(
            subject=subject, queue=queue, cb=cb, start_at=start_at
        )
        self._bus_handlers.append(bus_handler)

    async def subscribe_all(self, config: nats_conf.ScannerConfig) -> None:
        for bus_conf in config.subject_bus_configs:
            await self._subscribe(
                config=config,
                subject=bus_conf.subject,
                queue=bus_conf.queue,
                cb=functools.partial(
                    callbacks.start_scan, state_reporter=self._state_reporter
                ),
                start_at="last_received",
                stream=bus_conf.queue,
            )

            logger.info("subscribing to %s", bus_conf.subject)
            return

    async def _create_bus_handler(
        self, config: nats_conf.ScannerConfig
    ) -> scanner_handler.BusHandler:
        bus_handler = scanner_handler.BusHandler(
            bus_url=config.bus_url,
            cluster_id=config.bus_cluster_id,
            name=config.bus_client_name,
        )
        return bus_handler


async def connect_nats(
    config: nats_conf.ScannerConfig,
    scanner_id: str,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
) -> ScanHandler:
    """connecting to nats.

    Args:
        config: The key to connect to ostorlab.
        scanner_id: The scanner identifier.
        state_reporter: The scanner state report.
    """
    try:
        logger.info("starting bus runner for scanner %s", scanner_id)
        scan_handler = ScanHandler(state_reporter)
        logger.info("connected, subscribing to plans channels ...")
        await scan_handler.subscribe_all(config)
        logger.info("subscribed")
        return scan_handler
    except errors.ServiceUnavailableError as e:
        logger.exception("Failed to establish connection to NATs: %s", e)


async def subscribe_nats(
    api_key: str,
    scanner_id: str,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
) -> None:
    """Fetching the scanner configuration and subscribing to nats.

    Args:
        api_key: The key to connect to ostorlab.
        scanner_id: The scanner identifier.
        state_reporter: The scanner state report.
    """
    logger.info("Fetching scanner configuration.")
    runner = authenticated_runner.AuthenticatedAPIRunner(api_key=api_key)
    data = runner.execute(scanner_config.ScannerConfigAPIRequest(scanner_id=scanner_id))
    config = nats_conf.ScannerConfig.from_json(data)

    if config is None:
        logger.error("No config found to start the connection.")
        return

    logger.info("Connecting to nats.")
    await connect_nats(config, scanner_id, state_reporter)



"""Defines call back to trigger a scan after receiving a startAgentScan messages in the NATS."""
import logging
import ipaddress
from typing import List, Any

from ostorlab.runtimes import registry
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.cli import install_agent
from ostorlab.assets import asset
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.assets import android_store
from ostorlab.assets import android_aab
from ostorlab.assets import android_apk
from ostorlab.assets import file
from ostorlab.assets import ios_ipa
from ostorlab.assets import domain_name
from ostorlab.assets import link as link_asset
from ostorlab.assets import ios_store
from ostorlab.assets import agent as agent_asset
from ostorlab.utils import scanner_state_reporter


logger = logging.getLogger(__name__)


def _install_agents(runtime_instance: runtime.Runtime, agents) -> None:
    """Trigger installation of the agents that will run the scan."""
    try:
        runtime_instance.install()
        for agent in agents:
            install_agent.install(agent.key, agent.version)
    except install_agent.AgentDetailsNotFound:
        logger.warning("agent %s not found on the store", agent.key)


def _prepare_ip_asset(ip_request) -> asset.Asset:
    """Return IP assets from a NATs received message."""
    ip_network = ipaddress.ip_network(ip_request.host, strict=False)
    if ip_network.version == 4:
        return ipv4.IPv4(
            host=ip_network.network_address.exploded,
            mask=ip_request.mask or str(ip_network.prefixlen),
        )
    elif ip_network.version == 6:
        return ipv6.IPv6(
            host=ip_network.network_address.exploded,
            mask=ip_request.mask or str(ip_network.prefixlen),
        )
    else:
        raise ValueError(f"Invalid Ip address {ip_request.host}")


def _extract_assets(request: Any) -> List[asset.Asset]:
    """Returns list of specific Ostorlab-injectable assets, from a message received from NATs."""
    assets = []
    asset_type = request.WhichOneof("asset")
    if asset_type in ("ip", "ip4v", "ipv6"):
        ip_request = request.ip or request.ipv4 or request.ipv6
        assets.append(_prepare_ip_asset(ip_request))

    elif asset_type == "android_store":
        assets.append(
            android_store.AndroidStore(package_name=request.android_store.package_name)
        )

    elif asset_type == "ios_store":
        assets.append(ios_store.IOSStore(bundle_id=request.ios_store.bundle_id))

    elif asset_type == "ipa":
        assets.append(
            ios_ipa.IOSIpa(
                content=request.ipa.content,
                path=request.ipa.path,
                content_url=request.ipa.content_url,
            )
        )
    elif asset_type == "apk":
        assets.append(
            android_apk.AndroidApk(
                content=request.apk.content,
                path=request.apk.path,
                content_url=request.apk.content_url,
            )
        )
    elif asset_type == "aab":
        assets.append(
            android_aab.AndroidAab(
                content=request.aab.content,
                path=request.aab.path,
                content_url=request.aab.content_url,
            )
        )
    elif asset_type == "domain":
        assets.append(domain_name.DomainName(name=request.domain_name.name))

    elif asset_type == "agent":
        assets.append(
            agent_asset.Agent(
                key=request.agent.key,
                version=request.agent.version,
                git_location=request.agent.git_location,
                docker_location=request.agent.docker_location,
                yaml_file_location=request.agent.yaml_file_location,
            )
        )

    elif asset_type == "file":
        assets.append(
            file.File(
                content=request.file.content,
                path=request.file.path,
                content_url=request.file.content_url,
            )
        )
    elif asset_type == "network":
        for ip in request.network.ips:
            ip_asset = _prepare_ip_asset(ip)
            assets.append(ip_asset)

    elif asset_type == "links":
        for link in request.links.links:
            assets.append(
                link_asset.Link(
                    url=link.url,
                    method=link.url or "GET",
                )
            )

    else:
        logger.error(
            "%s not supported from scan reference  %s ",
            asset_type,
            request.reference_scan_id,
        )

    return assets


def _extract_agent_group_definition(request: Any) -> definitions.AgentGroupDefinition:
    return definitions.AgentGroupDefinition.from_bus_message(request)


def _extract_scan_id(request: Any) -> int:
    return request.scan_id


def _update_state_reporter(
    state_reporter: scanner_state_reporter.ScannerStateReporter, scan_id: int
) -> scanner_state_reporter.ScannerStateReporter:
    state_reporter.scan_id = scan_id
    return state_reporter


def start_scan(
    subject: str,
    request: Any,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
) -> None:
    """Responsible for triggering an Ostorlab scan, after receiving a startAgentScan message in NATs.

    Args:
        request: deserialized message.
        state_reporter: State reporter instance responsible for sending current state of the scanner.
    """
    logger.debug("Triggering scan after receiving message on: %s", subject)

    agent_group_definition = _extract_agent_group_definition(request)
    assets = _extract_assets(request)
    scan_id = _extract_scan_id(request)

    state_reporter = _update_state_reporter(state_reporter, request)

    runtime_instance = registry.select_runtime(
        runtime_type="local", scan_id=scan_id, run_default_agents=False
    )

    if runtime_instance.can_run(agent_group_definition=agent_group_definition) is True:
        _install_agents(runtime_instance, request.agents)

        runtime_instance.scan(
            agent_group_definition=agent_group_definition,
            assets=assets,
            title=None,
        )
    else:
        logger.error(
            "The runtime does not support the provided agent list or group definition."
        )
